'''Convenience methods and data for plotting.'''

import contextlib
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.animation as anim
import matplotlib.patches as patch
import numpy as np

from mpl_toolkits.mplot3d import Axes3D

# import many names directly from matplotlib.pyplot.
from matplotlib.pyplot import *

# colors from https://github.com/mbostock/d3/wiki/Ordinal-Scales
COLOR11 = (
    '#111111',
    '#d62728',
    '#1f77b4',
    '#2ca02c',
    '#9467bd',
    '#ff7f0e',
    '#e377c2',
    '#8c564b',
    '#bcbd22',
    '#7f7f7f',
    '#17becf',
    )


def plot(*args, **kwargs):
    '''Helper function for making quick line plots.

    This helper generates a new axes object and calls `matplotlib.pyplot.plot`.

    Positional and keyword arguments are passed directly to
    `matplotlib.pyplot.plot`.

    Returns
    -------
    matplotlib.pyplot.Axes :
        An axes object containing the plot.
    '''
    ax = create_axes(spines=True)
    ax.plot(*args, **kwargs)
    return ax


def scatter(*args, **kwargs):
    '''Helper function for making quick scatter plots.

    This helper generates a new axes object and calls `matplotlib.pyplot.scatter`.

    Positional and keyword arguments are passed directly to
    `matplotlib.pyplot.scatter`.

    Returns
    -------
    matplotlib.pyplot.Axes :
        An axes object containing the plot.
    '''
    ax = create_axes(spines=None)
    ax.scatter(*args, **kwargs)
    return ax


@contextlib.contextmanager
def axes(*args, **kwargs):
    '''Produce a 2D plotting axes, for use in a with statement.

    Parameters
    ----------
    show_afterwards : bool, optional
        If present and False, do not show the plot after the yield.

    All other arguments are passed to :func:`create_axes`.
    '''
    if 'show_afterwards' in kwargs:
        show_afterwards = kwargs.pop('show_afterwards')
    ax = create_axes(*args, **kwargs)
    yield ax
    if show_afterwards:
        show()


@contextlib.contextmanager
def axes3d(*args, **kwargs):
    '''Produce a 3D plotting axes, for use in a with statement.

    Parameters
    ----------
    show_afterwards : bool, optional
        If present and False, do not show the plot after the yield.

    All other arguments are passed to :func:`create_axes`.
    '''
    if 'show_afterwards' in kwargs:
        show_afterwards = kwargs.pop('show_afterwards')
    ax = create_axes(*args, projection='3d', spines=None, **kwargs)
    ax.w_xaxis.set_pane_color((1, 1, 1, 1))
    ax.w_yaxis.set_pane_color((1, 1, 1, 1))
    ax.w_zaxis.set_pane_color((1, 1, 1, 1))
    yield ax
    if show_afterwards:
        show()


def create_axes(subplot=111, shape=None, log='', max_xticks=5, max_yticks=4, spines=True, **kwargs):
    '''Create an axes object to hold a plot.

    Parameters
    ----------
    subplot : int or tuple of ints
        This will be passed to `matplotlib.pyplot.subplot()`.
    shape : tuple of ints
        If this is provided, it will be passed as the argument to
        `matplotlib.pyplot.axis()`.
    log : str
        A string, containing 'x' and/or 'y', indicating the axes that should
        have log scales.
    max_xticks : int, optional
        The maximum number of ticks to position on the x-axis, if any.
    max_yticks : int, optional
        The maximum number of ticks to position on the y-axis, if any.
    spines : any, optional
        If given, this argument will be passed to `set_spines`.

    Remaining keyword arguments will be passed to `axes()` (if `shape` is not
    None) or `subplot()` (otherwise).

    Returns
    -------
    matplotlib.Axes :
        A matplotlib.Axes instance.
    '''
    if shape is not None:
        ax = plt.axes(shape, **kwargs)
    else:
        if not isinstance(subplot, (tuple, list)):
            subplot = (subplot, )
        ax = plt.subplot(*subplot, **kwargs)

    if 'x' in log:
        ax.set_xscale('log')
    if 'y' in log:
        ax.set_yscale('log')

    if max_xticks > 0:
        plt.locator_params(axis='x', nbins=max_xticks - 1)
    if max_yticks > 0:
        plt.locator_params(axis='y', nbins=max_yticks - 1)

    if spines is not None:
        set_spines(ax, spines)

    return ax


def set_spines(ax, spines=True, offset=6):
    '''Set the spines on the given Axes object.

    Arguments
    ---------
    ax : matplotlib.Axes
        Set the spines on this Axes object.
    spines : any
        This parameter provides formatting instructions for the axes in the plot.
        It can take many different types of values:

        - dictionary: The strings 'left', 'right', 'top', 'bottom' should be
          mapped to numeric values---the numeric value specifies the distance
          outward from the plot (in points) to show the corresponding spine.
          Alternatively, the values in the dictionary can be anything that can
          be passed to `matplotlib.Spine#set_position`.

        - numeric: All spines will be shown, offset outward by this many points.

        - string/list/tuple/set: Spines with names included in the
          string/list/tuple/set will be shown at an offset of 6pt. Spines not
          named will not be shown.

        - falsey: No spines will be shown.

        Otherwise, the default spine configuration is used---the bottom and left
        spines are shown, offset outward by 6pt, and the top and right spines
        are hidden.
    offset : numeric
        Default offset for spine names mentioned in the `spines` argument.
    '''
    if spines is True:
        spines = 'bottom left'
    if spines is False:
        spines = ''

    if isinstance(spines, (int, float)):
        offset = spines
        spines = 'top right bottom left'

    if isinstance(spines, str):
        spines = spines.split()

    if isinstance(spines, (list, tuple, set)):
        spines = {k: offset for k in spines}

    for name, spine in ax.spines.items():
        pos = spines.get(name)
        if pos is None:
            spine.set_visible(False)
        elif isinstance(pos, int):
            spine.set_position(('outward', pos))
        else:
            spine.set_position(pos)

    if 'bottom' in spines:
        ax.xaxis.tick_bottom()
    else:
        ax.xaxis.set_ticks([])
    if 'left' in spines:
        ax.yaxis.tick_left()
    else:
        ax.yaxis.set_ticks([])


N = 127
t1 = np.linspace(0, 2 * np.pi, N)
t2 = np.linspace(0, np.pi, N)
XGRID = np.outer(np.cos(t1), np.sin(t2)).flatten()
YGRID = np.outer(np.sin(t1), np.sin(t2)).flatten()
ZGRID = np.outer(np.ones_like(t1), np.cos(t2)).flatten()


def eigs(ax, means, vals, vecs, ix=0, iy=1, contours=(1, ), **kwargs):
    '''Draw an ellipse at each contour.

    Parameters
    ----------
    ax : Axis3D
        Axis to plot on. This must be a 3D axis.
    means : ndarray (dims, )
        Means for each dimension of the data.
    vals : ndarray (dims, )
        Eigenvalues of the data.
    vecs : ndarray (dims, dims)
        Eigenvectors of the data. Each column is one eigenvector.
    ix : int, optional
        X-axis index. Defaults to 0.
    iy : int, optional
        Y-axis index. Defaults to 1.
    contours : sequence of float, optional
        List of contour lines to draw. Defaults to (1, ), i.e., just draw one
        contour at the scale of the eigenvalues.

    Returns
    -------
    surfaces :
        A list of the artists that were plotted.
    '''
    artists = []
    for c in contours:
        artists.append(patch.Ellipse(
            xy=[means[ix], means[iy]],
            width=c * vals[ix],
            height=c * vals[iy],
            angle=np.degrees(np.arctan2(vecs[iy, ix], vecs[ix, ix])),
            **kwargs))
    [ax.add_artist(art) for art in artists]
    return artists


def eigs3d(ax, means, vals, vecs, ix=0, iy=1, iz=2, contours=(1, ), **kwargs):
    '''Draw a 3D ellipse at each contour.

    Parameters
    ----------
    ax : Axis3D
        Axis to plot on. This must be a 3D axis.
    means : ndarray (dims, )
        Means for each dimension of the data.
    vals : ndarray (dims, )
        Eigenvalues of the data.
    vecs : ndarray (dims, dims)
        Eigenvectors of the data. Each column is one eigenvector.
    ix : int, optional
        X-axis index. Defaults to 0.
    iy : int, optional
        Y-axis index. Defaults to 1.
    iz : int, optional
        Z-axis index. Defaults to 2.
    contours : sequence of float, optional
        List of contour lines to draw. Defaults to (1, ), i.e., just draw one
        contour at the scale of the eigenvalues.

    Returns
    -------
    surfaces :
        A list of the surfaces that were plotted.
    '''
    r = np.dot(vecs[[ix, iy, iz], [ix, iy, iz]],
               np.vstack([vals[ix] * XGRID, vals[iy] * YGRID, vals[iz] * ZGRID]))
    surfaces = []
    for c in contours:
        surfaces.append(ax.plot_surface(
            means[ix] + c * r[ix].reshape((N, N)),
            means[iy] + c * r[iy].reshape((N, N)),
            means[iz] + c * r[iz].reshape((N, N)), **kwargs))
    return surfaces
